/* Global declarations file for text formatter */

ushort REVISION_LEVEL = 7;

int VERSION = 075;

bool TESTMODE = false;

int DEFAULTPAGEWIDTH = 80,	/* Default physical page width */
    DEFAULTPAGELENGTH = 66,	/* Default physical page length */
    HUGE = 9999;		/* A big number */

char
    DEFAULTCOMMANDCHAR = '.',	/* Command lines start with a period */
    DEFAULTSPECIALCHAR = '\#',	/* Escape character for exprs in text lines */
    NONTRIVIALBLANKCHAR = '~',	/* Non-trivial blank */
    NORMALSPACINGCHAR = 'n',	/* Used in regular flush units */
    LINESPACINGCHAR = 'l',	/* Used in line-spacing flush units */
    BOLDFACEFLAG = 'b', 	/* Start/stop bold-facing (#b/#-b) */
    UNDERLINEFLAG = 'u';	/* Start/stop underlining (#u/#-u) */

byte
    T_UNDERLINE = 0x80, 	/* Underline this character */
    T_BOLDFACE = 0x40,		/* Boldface this character */
    T_SPECIAL = 0x20,		/* This character is a flag character */
    T_FREE = 0x01;		/* This is a free TEXT object */

/* Flag bits used in CharMasks vector to indicate char type */
byte
    C_UNDERLINE = T_UNDERLINE,	/* Underline this character on output */
				/* Note:  setCharFlags assumes that */
				/*	C_UNDERLINE = T_UNDERLINE */
    C_VARNAME = 0x40;		/* Legal character in variable names */

type
    
    DEVUNIT = int,		/* Type used to contain a device unit */
    
    TEXTUNIT = struct { 	/* Objects used in vectors of text objects */
	char tu_ch;
	byte tu_flags
    },

    TEXT = struct {		/* A text unit (used to hold a text char) */
	char t_ch;		/* The character */
	*TEXT t_next;		/* The pointer to the next TEXT unit */
	byte t_flags		/* The flag byte (see T_... above) */
    },

    BUFFER = struct {		/* Object used to hold lists of text objects */
	*TEXT b_start;		/* Pointer to first text object */
	*TEXT b_end;		/* Pointer to last text object */
    },

    COMMANDTYPE = enum{
	C_UNKNOWN,
	C_SPACE,		/* space down page	*/
	C_FLUSH,		/* flush output unit	*/
	C_SOURCE,		/* source input file	*/
	C_MACLIB,		/* source system macros */
	C_MACRO,		/* define macro 	*/
	C_COMMANDMACRO, 	/* define a 'command'	*/
	C_PRINT,		/* print message	*/
	C_NUMBER,		/* define number vars	*/
	C_STRING,		/* define string vars	*/
	C_TEXT, 		/* define text vars	*/
	C_DIVERSION,		/* define diversions	*/
	C_LNUMBER,		/* define local numbers */
	C_LSTRING,		/* define local strings */
	C_LTEXT,		/* define local texts	*/
	C_LDIVERSION,		/* define local div.	*/
	C_FNUMBER,		/* define file number	*/
	C_FSTRING,		/* define file string	*/
	C_FDIVERSION,		/* define file div.	*/
	C_DESTROY,		/* destroy a symbol	*/
	C_DIVERT,		/* use a diversion	*/
	C_ENDDIVERT,		/* stop a diversion	*/
	C_INCLUDE,		/* include a diversion	*/
	C_PRINTSTATS,		/* print statistics	*/
	C_AT,			/* specify an at-point	*/
	C_CANCEL,		/* cancel an at-point	*/
	C_IF,			/* start an IF		*/
	C_ELSEIF,		/* ELIF part of IF	*/
	C_ELSE, 		/* ELSE part of IF	*/
	C_ENDIF,		/* end an IF		*/
	C_WHILE,		/* start a WHILE loop	*/
	C_ENDWHILE,		/* end body of WHILE	*/
	C_PUT,			/* put line to output	*/
	C_SINK, 		/* change output files	*/
	C_PUSH, 		/* Push a variable	*/
	C_POP,			/* Pop a variable	*/
	C_FPUSH,		/* Push a file variable */
	C_FPOP, 		/* Pop a file variable	*/
	C_TIC,			/* tic tok		*/
    },

    CHARSLIST = struct {	/* Object used to build lists of strings */
	*CHARSLIST cl_next;	/* Pointer to next element */
	*char cl_this		/* The string (*char) */
    },

    TEXTSLIST = struct {	/* Object used to hold list of text vectors */
	*TEXTSLIST tl_next;	/* Pointer to next element */
	*TEXTUNIT tl_this;	/* Pointer to text vector (terminated by \e) */
	int tl_offset		/* don't remember what this is for... */
    },
 
    ATLIST = struct {		/* Object used to hold list of AT points */
	*ATLIST at_next;	/* Pointer to next element */
	int at_line;		/* What line this AT point is for */
	*char at_command;	/* text to include when AT is triggered */
	*char at_name		/* Name of this AT point */
    },

    SINKSTATUS = struct {
	BUFFER ss_outBuf;
	bool ss_spreadDirection;/* direction of line spread (alternates) */
	bool ss_noTextYetThisPage; /* Is there any text on this page yet */
	int ss_lineNumber;	/* number of current line */
	int ss_wordCount;	/* # words currently in output line */
	int ss_justify; 	/* 0 - not justifying, ~0 - justifying */
	int ss_fill;		/* 0 - not filling, ~0 - filling lines */
	int ss_flushOK; 	/* 0 - flush command disabled */
	int ss_lineSpacing;	/* spacing between output lines */
	int ss_leftIndent;	/* current left indent value */
	int ss_rightIndent;	/* current right indent value */
	int ss_tempIndent;	/* temporary indent for current line */
	int ss_actualPPL;	/* Actual Physical Page Length */
	int ss_actualPPW;	/* Actual Physical Page Width */
	int ss_PPL;		/* Physical Page Length */
	int ss_PPW;		/* Physical Page Width */
	int ss_actualTM;	/* Actual Top Margin */
	int ss_actualBM;	/* Actual Bottom Margin */
	int ss_actualLM;	/* Actual Left Margin */
	int ss_actualRM;	/* Actual Right Margin */
	int ss_topMargin;	/* Top Margin */
	int ss_bottomMargin;	/* Bottom Margin */
	int ss_leftMargin;	/* Left Margin */
	int ss_rightMargin;	/* Right Margin */
	int ss_actualPL;	/* Actual Page Length */
	int ss_actualPW;	/* Actual Page Width */
	int ss_currentPage;	/* Current Page Number */
	int ss_nextPage;	/* Next Page Number */
	*CHARSLIST ss_atQueueHead; /* Head of queued at conditions */
	*CHARSLIST ss_atQueueTail; /* Tail of ... */
	bool ss_atsEnabled;	/* Are at conditions currently enabled? */
	*ATLIST ss_topAts;	/* at's relative to top of page */
	*ATLIST ss_bottomAts;	/* at's relative to bottom of page */
	*ATLIST ss_eofAts;	/* at's at end-of-file */
	**ATLIST ss_nextTopAt;	/* next top at to be triggered */
	**ATLIST ss_nextBottomAt;/* next bottom at to be triggered */
    },

    DIVERSIONTYPE = enum {	/* Type of diversions possible */
	D_FILE, 		/* File diversion (not supported) */
	D_MEMORY,		/* Memory (regular) diversion */
	D_BITBUCKET		/* *DUMMY* diversion */
    },

    DIVERSION = struct {	/* Structure used to describe a diversion */
	DIVERSIONTYPE d_type;	/* Type of this diversion */
	bool d_inUse;		/* Is this diversion being included? */
	int d_lineCount;	/* Number of lines in the diversion */
	SINKSTATUS d_status;	/* Sink-related information about diversion */
	*TEXTSLIST d_start, d_end	/* List of lines in the diversion */
    },

    DIVERSIONLIST = struct {	/* Object used to hold lists of diversions */
	*DIVERSIONLIST dl_next; /* Pointer to next diversion list element */
	*DIVERSION dl_this;	/* Pointer to this diversion */
	int dl_count;		/* Number of lines to include from this one */
    },

    MACRO = struct {		/* Object used to hold a macro */
	int m_useCount; 	/* # of times macro has been called and not */
				/*	yet returned */
	*CHARSLIST m_body	/* Pointer to text of macro */
    },

    FLUSHTYPE = enum {		/* Types of flush units */
	FL_INFINITE,		/* Infinite (page eject) */
	FL_NORMAL,		/* Normal (specific line count) */
	FL_LINESPACING		/* Line spacing (generate linespacing lines) */
    },

    FLUSHUNIT = struct {	/* Object to hold a flush unit */
	*FLUSHUNIT fl_next;	/* Pointer to next flush unit on this page */
	**FLUSHUNIT fl_head;	/* Pointer to first flush unit on this page */
	FLUSHTYPE fl_type;	/* Type of this flush unit */
	int fl_count		/* Number of lines left in this unit */
    },

    OBJECTTYPE = enum { 	/* Types of objects */
	OB_UNKNOWN,		/* Unknown (like not-a-number) */
	OB_COMMAND,		/* Command */
	OB_NUMBER,		/* Regular number (user defined or */
				/*	the result of an expression */
	OB_STRING,		/* Regular string */
	OB_TEXT,		/* Text object */
	OB_MACRO,		/* Macro */
	OB_COMMANDMACRO,	/* Command macro */
	OB_RONUMBER,		/* Read-only system number (points to number)*/
	OB_ROSTRING,		/* Read-only system string (points to string)*/
	OB_SYSTEMNUMBER,	/* System number (points to number) */
	OB_SYSTEMSTRING,	/* System string (points to string) */
	OB_DIVERSION,		/* Diversion */
	OB_FUNCTION,		/* Function */
    },

    OBJECTVALUE = union {	/* Union to hold the value of an object */
	COMMANDTYPE ob_command; /* Which command */
	int ob_number;		/* Regular number */
	*int ob_systemNumber;	/* Pointer to system number (RW and RO) */
	*char ob_string;	/* Regular string */
	BUFFER ob_text; 	/* Text */
	*MACRO ob_macro;	/* Pointer to macro */
	**char ob_systemString; /* Pointer to system string (RW and RO) */
	*DIVERSION ob_diversion;/* Pointer to diversion */
	proc()void ob_function	/* Address of function */
    },

    OBJECT = struct {		/* Structure to hold an object */
	OBJECTVALUE ob_value;	/* Value of object */
	OBJECTTYPE ob_type	/* Type of object */
    },

    SYMBOL = struct {		/* Structure to hold a symbol definition */
	*SYMBOL sy_next;	/* Pointer to next symbol in table */
	bool sy_visible;	/* Is this symbol visible (not pushed)? */
	*char sy_name;		/* Name of this symbol */
	OBJECT sy_value 	/* Object containing value of this symbol */
    };

BUFFER
    NameBuf,		/* temporary BUFFER used all over the place */
    InBuf;		/* input buffer for data and command lines */

int ActiveIfCount;	/* Number of active if's in current source unit */

int EchoInput;		/* User flag to force echoing of input */

bool SystemMode;	/* true if reading a system macro file */

[256]byte CharMasks;	/* Lookup table of character type bits */
